#version 430
#extension GL_ARB_separate_shader_objects : enable










struct SComputeInput
{
	uvec3 GlobalThreadID; uvec3 LocalThreadID; uvec3 GroupID;};

layout(std430 ) readonly buffer BStreamInfo
{
	uint StreamInfo[];
} StreamInfo;
layout(std430 ) readonly buffer BGPUSimData
{
	uint GPUSimData[];
} GPUSimData;
layout(std430 ) readonly buffer BEnabledOffsets
{
	uint EnabledOffsets[];
} EnabledOffsets;
layout(std430 ) buffer BOutIndirection
{
	uint OutIndirection[];
} OutIndirection;
layout(std430 ) buffer BIndirectionOffsets
{
	uint IndirectionOffsets[];
} IndirectionOffsets;
layout(std430 ) readonly buffer BMeshIDsOffsets
{
	uint MeshIDsOffsets[];
} MeshIDsOffsets;

layout( std140) uniform UInfo
{
	uint DrawRequest;
	uint MeshCount;
} Info;

void ComputeMain( in SComputeInput  cInput  )
{
	const uint particleID = cInput.GlobalThreadID.x;
	const uint count =  StreamInfo.StreamInfo [0] ;
	if (particleID >= count)
		return;

	const uint drawRequest =  Info.DrawRequest ;
	const uint enabledOffset =  EnabledOffsets.EnabledOffsets [ drawRequest ] ;
	const uint enabled =  GPUSimData.GPUSimData [enabledOffset + particleID ] ;
	if (enabled == 0)
		return;


	const uint meshCount =  Info.MeshCount ;
	const uint meshIDsOffset =  MeshIDsOffsets.MeshIDsOffsets [ drawRequest ] ;
	const uint meshID = min(uint( uintBitsToFloat(GPUSimData.GPUSimData [meshIDsOffset + particleID ]) ), meshCount - 1);
	const uint indirectionOffsetID = drawRequest * meshCount + meshID;
	uint outIndex;
	outIndex = atomicAdd( IndirectionOffsets.IndirectionOffsets [ indirectionOffsetID ], floatBitsToUint(4) ) ;

	OutIndirection.OutIndirection [outIndex] = particleID; ;
}

layout(local_size_x = 128, local_size_y = 1, local_size_z = 1) in;
void main()
{
	SComputeInput cInput;
	cInput.GlobalThreadID = gl_GlobalInvocationID;
	cInput.LocalThreadID = gl_LocalInvocationID;
	cInput.GroupID = gl_WorkGroupID;
	ComputeMain(cInput);
}
