#version 430
#extension GL_ARB_separate_shader_objects : enable





layout(location = 0) in vec2 _vin_TexCoords;

struct SVertexInput
{
	vec2 TexCoords;
	int VertexIndex;
	uint InstanceId;
};

layout(location = 0) smooth out vec4 _vout_fragColor;

struct SVertexOutput
{
	vec4 fragColor;
	vec4 VertexPosition;
};

layout( std140) uniform USceneInfo
{
	mat4 ViewProj;
	mat4 View;
	mat4 Proj;
	mat4 InvView;
	mat4 InvViewProj;
	mat4 UserToLHZ;
	mat4 LHZToUser;
	mat4 BillboardingView;
	mat4 PackNormalView;
	mat4 UnpackNormalView;
	vec4 SideVector;
	vec4 DepthVector;
	vec2 ZBufferLimits;
	float Handedness;
} SceneInfo;
layout( std140) uniform UBillboardInfo
{
	vec4 DrawRequest[256];
} BillboardInfo;
layout(std430 ) readonly buffer BIndices
{
	uint Indices[];
} Indices;
layout(std430 ) readonly buffer BPositions
{
	uint Positions[];
} Positions;
layout(std430 ) readonly buffer BSize2s
{
	uint Size2s[];
} Size2s;
layout(std430 ) readonly buffer BAxis0s
{
	uint Axis0s[];
} Axis0s;

layout( std140) uniform UGPUBillboardPushConstants
{
	uint IndicesOffset;
} GPUBillboardPushConstants;
vec2 get_radius(uint particleID  )
{
	float enabled = 1.0f;
	return  vec2( uintBitsToFloat(Size2s.Size2s [ particleID * 2 ]) , uintBitsToFloat(Size2s.Size2s [( particleID * 2 ) + 1]) )  * enabled;
}


void RotateTangents(uint particleID,  inout vec3  tangent0,  inout vec3  tangent1  )
{
}


vec4 proj_position( in vec3  position  )
{

	return  (( SceneInfo.ViewProj ) * (vec4(position, 1.0f))) ;
}


vec3 ComputeTangent1( in vec3  nrm,  in vec3  tangent0  )
{
	vec3 side =  SceneInfo.SideVector .xyz;
	vec3 depth =  SceneInfo.DepthVector .xyz;

	vec3 tangent1 =  cross(nrm, tangent0) * SceneInfo.Handedness ;
	if (dot(tangent1, tangent1) >  1.0e-8f )
		tangent1 = normalize(tangent1);
	else
	{
		const vec3 v = -side * dot(nrm, depth) + depth * (dot(nrm, side) + 0.01f);

		tangent1 = normalize(cross(nrm, v));


		tangent1 = -tangent1;
	}
	return tangent1;
}


void VertexBillboard( in SVertexInput  vInput,  inout SVertexOutput  vOutput, uint particleID  )
{





	const vec4 worldPosDrId =  vec4( uintBitsToFloat(Positions.Positions [ particleID * 4 ]) , uintBitsToFloat(Positions.Positions [( particleID * 4 ) + 1]) , uintBitsToFloat(Positions.Positions [( particleID * 4 ) + 2]) , uintBitsToFloat(Positions.Positions [( particleID * 4 ) + 3]) ) ;
	const vec3 worldPos = worldPosDrId.xyz;
	const uint drId =  floatBitsToUint(worldPosDrId.w) ;


	const vec4 currentDr =  BillboardInfo.DrawRequest [drId];
	const uint flags =  floatBitsToUint(currentDr.x) ;
	const uint billboarderType = flags &  7U ;

	const vec2 radius = get_radius(particleID  );
	vec3 tangent0 = vec3(0, 0, 0);
	vec3 tangent1 = vec3(0, 0, 0);
	vec2 texCoords = vInput.TexCoords;
	switch (billboarderType)
	{

	case  1U :
	{
		tangent0 =  SceneInfo.BillboardingView [0] .xyz * radius.x;
		tangent1 =  SceneInfo.BillboardingView [1] .xyz * radius.y;
		RotateTangents(particleID, tangent0, tangent1  );
		break;
	}


	case  2U :
	{
		const vec3 viewPos =  SceneInfo.BillboardingView [3] .xyz;
		const vec3 camToParticle = normalize(worldPos - viewPos);
		const vec3 upVector =  cross( SceneInfo.DepthVector .xyz, SceneInfo.SideVector .xyz) * SceneInfo.Handedness ;
		const vec3 upAxis = upVector + 0.01f * dot(upVector, camToParticle) *  SceneInfo.DepthVector .xyz;

		tangent0 = normalize( cross(camToParticle, upAxis) * SceneInfo.Handedness );
		tangent1 =  cross(tangent0, camToParticle) * SceneInfo.Handedness  * radius.y;
		tangent0 *= radius.x;
		RotateTangents(particleID, tangent0, tangent1  );
		break;
	}


	case  3U :
	{
		const vec3 viewPos =  SceneInfo.BillboardingView [3] .xyz;
		const vec3 camToParticle = normalize(worldPos - viewPos);




		const vec3 axis_fwd =  vec3( uintBitsToFloat(Axis0s.Axis0s [ particleID * 3 ]) , uintBitsToFloat(Axis0s.Axis0s [( particleID * 3 ) + 1]) , uintBitsToFloat(Axis0s.Axis0s [( particleID * 3 ) + 2]) ) ;


		tangent0 = ComputeTangent1(camToParticle, axis_fwd  ) * radius.x;
		tangent1 = axis_fwd * 0.5f;
		break;
	}


	case  4U :
	{
		const vec3 viewPos =  SceneInfo.BillboardingView [3] .xyz;
		const vec3 camToParticle = normalize(worldPos - viewPos);




		const vec3 axis_fwd =  vec3( uintBitsToFloat(Axis0s.Axis0s [ particleID * 3 ]) , uintBitsToFloat(Axis0s.Axis0s [( particleID * 3 ) + 1]) , uintBitsToFloat(Axis0s.Axis0s [( particleID * 3 ) + 2]) ) ;


		tangent0 = ComputeTangent1(camToParticle, axis_fwd  ) * radius.x;
		tangent1 = axis_fwd * 0.5f +  cross(tangent0, camToParticle) * SceneInfo.Handedness ;
		break;
	}
	default:
		break;
	}


	tangent1 *= -1.0f;


	const vec3 bbCorner = tangent0 * texCoords.x + tangent1 * texCoords.y;
	const vec3 vertexWorldPosition = worldPos + bbCorner;

	vOutput.VertexPosition = proj_position(vertexWorldPosition  );
}

void VertexMain( in SVertexInput  vInput,  out SVertexOutput  vOutput  )
{
	const uint indicesOffset =  GPUBillboardPushConstants.IndicesOffset ;




		const uint particleID =  Indices.Indices [ indicesOffset + vInput.InstanceId ] ;


	float degenerate = 1.0f;
	vOutput.fragColor = vec4(1.0, 0.0, 1.0, 1.0);
	VertexBillboard(vInput, vOutput, particleID  );

	vOutput.VertexPosition *= degenerate;
}


void main()
{
	SVertexInput vInput;
	SVertexOutput vOutput;

	vInput.TexCoords = _vin_TexCoords;
	vInput.VertexIndex = gl_VertexID;
	vInput.InstanceId = uint(gl_InstanceID);
	vOutput.fragColor = vec4(0, 0, 0, 0);
	VertexMain(vInput, vOutput);
	_vout_fragColor = vOutput.fragColor;
	gl_Position = vOutput.VertexPosition;
}
